#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# File:     main.py
# Author:   Daniel Dyringer
# Brief:    AIRTLS Anchor Shell client
# -----------------------------------------------------------------------------

from typing import List
import click
import click_shell
import traceback
import sys
import os
import logging
from datetime import datetime

logger = logging.getLogger(__name__)

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))  # noqa

from lib_acp_tlv import *


device_finder = DeviceManager()
device_instance: InterfaceDevice = None
device_connectable_interfaces: list = []


def handle_get_devices():
    devices = device_finder.get_all_devices()
    device_finder.print_devices_list()

    for did in devices.keys():
        for interface in devices[did]:
            if ("-USB-PROTOCOL" in interface.name) or ("-BLE" in interface.name):
                device_connectable_interfaces.append(f"{did}:{interface.name}")

    for d in device_connectable_interfaces:
        f = bind_connect(d)
        connect.command(name=d)(f)


@click.group(help="Connect to device")
def connect():
    pass


def bind_connect(device_str: str) -> None:

    def func():
        global device_finder
        global device_instance

        if device_instance != None:
            try:
                device_instance.disconnect()
            except Exception as e:
                pass

        did, type = device_str.split(":")
        device_finder.get_all_devices(ble=("BLE" in type), cdc=("USB" in type))
        device_instance = device_finder.get_device(did, type)
        device_instance.connect()

    return func


@click.command()
def get_devices():
    handle_get_devices()


@click.command()
def list_usb():
    device_finder.get_all_devices(ble=False, cdc=True)
    device_finder.print_devices_list()


@click.command()
def disconnect():
    global device_instance
    if device_instance != None:
        device_instance.disconnect()
    device_instance = None


@click.command()
def dump_frames():
    global device_finder
    frames = device_finder.get_all_frames()
    for ts, frame in frames:
        print(f"RX Timestamp: {datetime.fromtimestamp(ts)}")
        print(f"Data:{frame}\n")


@click.command()
def logging_enable():
    global device_finder
    device_finder.enable_logging(True)


@click.command()
def logging_disable():
    global device_finder
    device_finder.enable_logging(False)


@click.command()
@click.argument('pattern', type=click.IntRange(min=0, max=63))
@click.argument('brightness', type=click.IntRange(min=0, max=3))
@click.argument('cycle_id', type=click.IntRange(min=0, max=65535))
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
def cmd_blink(pattern: int, brightness: int, cycle_id: int, did1, did2):
    global device_instance
    if device_instance != None:
        try:
            cycle_id = cycle_id & 0xFFFF
            
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"
                
            data = {
                "anchors": [
                    {
                        "did": did1,
                        "pattern": pattern,
                        "brightness" : brightness,
                        "cycle_id": cycle_id,
                    },
                    {
                        "did": did2,
                        "pattern": pattern,
                        "brightness" : brightness,
                        "cycle_id": cycle_id,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandBlinkFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")

@click.command()
@click.argument('pattern', type=click.IntRange(min=0, max=63))
@click.argument('brightness', type=click.IntRange(min=0, max=3))
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
@click.argument('did3', type=str, required=False)
def cmd_blink_immediately(pattern: int, brightness: int, did1, did2, did3):
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"
            if did3 == None:
                did3 = "000000"

            data = {
                "anchors": [
                    {
                        "did": did1,
                        "pattern": pattern,
                        "brightness" : brightness,
                    },
                    {
                        "did": did2,
                        "pattern": pattern,
                        "brightness" : brightness,
                    },
                    {
                        "did": did3,
                        "pattern": pattern,
                        "brightness" : brightness,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandBlinkImmediatelyFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")


@click.command()
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
@click.argument('did3', type=str, required=False)
@click.argument('did4', type=str, required=False)
def cmd_shutdown(did1, did2, did3, did4):
    """
    \b
    Turns off the devices 
    Arguments: 
     - did1 - optional device DID. If skipped, master anchor DID 
     - did2 - optional device DID 
     - did3 - optional device DID 
     - did4 - optional device DID 
    """
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"
            if did3 == None:
                did3 = "000000"
            if did4 == None:
                did4 = "000000"

            data = {
                "anchors": [
                    {
                        "did": did1,
                    },
                    {
                        "did": did2,
                    },
                    {
                        "did": did3,
                    },
                    {
                        "did": did4,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandShutdownFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")


@click.command()
@click.argument('ch3', type=click.IntRange(min=-128, max=127))
@click.argument('ch5', type=click.IntRange(min=-128, max=127))
@click.argument('txpower', type=click.IntRange(min=0, max=255))
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
def cmd_set_uwb_config(ch3: int, ch5: int, txpower: int, did1, did2):
    """
    \b
    Sets UWB config
    Arguments: 
     - ch3 - antenna delay offset for channel 3
     - ch5 - antenna delay offset for channel 5
     - txPower - TX power preset (not implemented yet)
     - did1 - optional device DID. If skipped, master anchor DID 
     - did2 - optional device DID
    """
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"

            data = {
                "devices": [
                    {
                        "did": did1,
                        "ch3": ch3,
                        "ch5": ch5,
                        "txPower": txpower,
                    },
                    {
                        "did": did2,
                        "ch3": ch3,
                        "ch5": ch5,
                        "txPower": txpower,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandSetUWBConfigFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")

@click.command()
@click.argument('config', type=click.IntRange(min=-128, max=127))
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
@click.argument('did3', type=str, required=False)
def cmd_set_board_config(config: int, did1, did2, did3):
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"
            if did3 == None:
                did3 = "000000"

            data = {
                "devices": [
                    {
                        "did": did1,
                        "config": config,
                    },
                    {
                        "did": did2,
                        "config": config,
                    },
                    {
                        "did": did3,
                        "config": config,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandSetBoardConfigFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")

@click.command()
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
@click.argument('did3', type=str, required=False)
@click.argument('did4', type=str, required=False)
def cmd_factory_reset(did1, did2, did3, did4):
    """
    \b
    Factory resets the device 
    Arguments: 
     - did1 - optional device DID. If skipped, master anchor DID 
     - did2 - optional device DID 
     - did3 - optional device DID 
     - did4 - optional device DID 
    """
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"
            if did3 == None:
                did3 = "000000"
            if did4 == None:
                did4 = "000000"

            data = {
                "anchors": [
                    {
                        "did": did1,
                    },
                    {
                        "did": did2,
                    },
                    {
                        "did": did3,
                    },
                    {
                        "did": did4,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandFactoryResetFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")

@click.command()
@click.argument('primary_seat', type=click.IntRange(min=0, max=255))
@click.argument('frequency', type=click.IntRange(min=0, max=9))
@click.argument('allocator_type', type=click.IntRange(min=0, max=2))
@click.argument('device_type', type=click.IntRange(min=0, max=4))
@click.argument('did1', type=str, required=False)
def cmd_set_network_config(primary_seat, frequency, allocator_type, device_type, did1):
    """
    \b
    Sets Network device config
    Args:
    - primary_seat - absolute seat number
    - frequency:
        0 - Unknown      (should not be used),
        1 - Random Tag   (not implemented),
        2 - VerySlow Tag (not implemented),
        3 - Slow Tag     (not implemented),
        4 - 25Hz,
        5 - 50Hz,
        6 - 75Hz,   (not implemented)
        7 - 100Hz,  (not implemented)
        8 - 125Hz,  (not implemented)
        9 - 150Hz,
    - allocator_type:
        0 - static
        1 - dynamic strict
        2 - dynamic relaxed
    - device_type:
        0 - None    (should not be used),
        1 - Dummy   (should not be used),
        2 - Anchor,
        3 - Tag,
        4 - Sniffer,
    """
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did

            data = {
                "devices": [
                {
                    "did": did1,
                    "primary_seat": primary_seat,
                    "frequency": frequency,
                    "allocator_type": allocator_type,
                    "device_type": device_type,
                },
                {
                    "did": "000000",
                    "primary_seat": 0x00,
                    "frequency": 0x00,
                    "allocator_type": False,
                    "device_type": 0,
                },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandSetNetworkConfigFrame, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")


@click.command()
@click.argument('period', type=int, required=True)
@click.argument('did1', type=str, required=False)
@click.argument('did2', type=str, required=False)
def cmd_set_random_period(period: int, did1: str, did2: str):
    """
    \b
    Factory resets the device 
    Arguments: 
     - period - random frequency tag active period in seconds
     - did1 - optional device DID. If skipped, master anchor DID 
     - did2 - optional device DID 
    """
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"

            data = {
                "devices": [
                    {
                        "did": did1,
                        "period": period,
                    },
                    {
                        "did": did2,
                        "period": period,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandSetRandomPeriod, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")


@click.command()
@click.argument("delay", type=int, required=True)
@click.argument("did1", type=str, required=False)
@click.argument("did2", type=str, required=False)
@click.argument("did3", type=str, required=False)
def cmd_set_network_start_delay(delay: int, did1: str, did2: str, did3: str):
    """
    \b
    Factory resets the device
    Arguments:
     - delay - anchor network start delay in seconds
     - did1 - optional device DID. If skipped, master anchor DID
     - did2 - optional device DID
     - did3 - optional device DID
    """
    global device_instance
    if device_instance != None:
        try:
            if did1 == None:
                did1 = device_instance.did
            if did2 == None:
                did2 = "000000"
            if did3 == None:
                did3 = "000000"

            data = {
                "devices": [
                    {
                        "did": did1,
                        "delay": delay,
                    },
                    {
                        "did": did2,
                        "delay": delay,
                    },
                    {
                        "did": did3,
                        "delay": delay,
                    },
                ]
            }

            frame = protocol_helper_createBinaryFrame(CommandSetNetworkStartDelay, data)
            device_instance.transmit(frame)

        except Exception as e:
            logger.error(e)
            print(traceback.format_exc())

    else:
        logger.error("Connect to the device first!")


def cleanup(_):
    global device_instance
    if device_instance != None:
        device_instance.disconnect()


@click_shell.shell(prompt='> ', intro='AIRTLS Shell', on_finished=cleanup)
def cli():
    handle_get_devices()


cli.add_command(get_devices)
cli.add_command(list_usb)
cli.add_command(connect)
cli.add_command(disconnect)
cli.add_command(dump_frames)
cli.add_command(logging_enable)
cli.add_command(logging_disable)
cli.add_command(cmd_blink)
cli.add_command(cmd_blink_immediately)
cli.add_command(cmd_shutdown)
cli.add_command(cmd_set_uwb_config)
cli.add_command(cmd_set_board_config)
cli.add_command(cmd_set_network_config)
cli.add_command(cmd_factory_reset)
cli.add_command(cmd_set_random_period)
cli.add_command(cmd_set_network_start_delay)

if __name__ == '__main__':
    cli()