# Receives frames from UDP socket and prints them as JSON
# Input arguments:
#   '-t', '--type' - optional ACP TLV frame type filter. Where type:
#       1 :  TimeOfFlightFrame,
#       2 :  SerialDataFrame,
#       3 :  AdminDistributionInfoFrame,
#       4 :  CompressedDataFrame,
#       5 :  AccelerometerDataFrame,
#       6 :  GyroscopeDataFrame,
#       7 :  MagnetometerFrame,
#       8 :  ImpactFrame,
#       
#       Example - print only accelerometer data frames:
#       python3 ./software/app_socket_receiver/main.py -t 5
#
#   '-d', '--did' - optional device ID filter.
#       Example - print only frames from device 'c2e151':
#       python3 ./software/app_socket_receiver/main.py -d c2e151
#

import socket
import os
import json
import argparse

UDP_IP = "127.0.0.1"
UDP_PORT = 5005

class SocketReceiver():
    def __init__(self, type_filter = None, did_filter = None) -> None:
        self.socket_instance = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket_instance.bind((UDP_IP, UDP_PORT))

        self.did_filter = did_filter
        self.type_filter = None
        if type_filter != None:
            self.type_filter = int(type_filter)

        self.type_frame_counters = {}
        self.did_frame_counters = {}

    def receive(self):
        data, addr = self.socket_instance.recvfrom(2048)
        if (data != None):
            data = data.decode("utf-8")
            try:
                # Decoding
                tlv_frame = json.loads(data)
                frame_type = tlv_frame["header"]["type"]
                frame_did = tlv_frame["value"]["did"]

                # Counters 
                if frame_type not in self.type_frame_counters.keys():
                    self.type_frame_counters[frame_type] = 1
                else:
                    self.type_frame_counters[frame_type] += 1

                if frame_did not in self.did_frame_counters.keys():
                    self.did_frame_counters[frame_did] = 1
                else:
                    self.did_frame_counters[frame_did] += 1

                # Filters
                if self.type_filter != None:
                    if self.type_filter != frame_type:
                        data = None

                if self.did_filter != None:
                    if self.did_filter != frame_did:
                        data = None
            except:
                pass

        return data

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--type', type=int, required=False)
    parser.add_argument('-d', '--did', type=str, required=False)
    args = parser.parse_args()
    type_filter = args.type
    did_filter = args.did

    client = SocketReceiver(type_filter, did_filter)

    while True:
        message = client.receive()
        if message != None:
            os.system('cls' if os.name == 'nt' else 'clear')
            print("Frame type counters:", client.type_frame_counters)
            print("DID frame counters:", client.did_frame_counters)
            print(message)