# -----------------------------------------------------------------------------
# File:     device.py
# Author:   Daniel Dyringer
# Brief:    Implements device manager class
# -----------------------------------------------------------------------------

import time
import logging

from .interface.abstract_interface import *
from .interface.ble import InterfaceBLE
from .interface.cdc import InterfaceSerial
from .protocol  import *
from .protocol_builder import ProtocolBuilder
from .socket_transmitter import SocketTransmitter

logger = logging.getLogger(__name__)

class DeviceManager():
    def __init__(self) -> None:
        self.protocol_handler = ProtocolBuilder(self._on_frame_callback)
        self.ble_interface = InterfaceBLE(self.protocol_handler.on_data)
        self.serial_interface = InterfaceSerial(self.protocol_handler.on_data)
        self.devices = {}
        self.frames = []
        self.last_measurement_ts = time.time()
        self.cycle_id = 0
        self.offload = SocketTransmitter()

    def _on_frame_callback(self, frame: dict)-> None:
        self.frames.append((time.time(), frame))
        self.offload.push(frame)
        if "cycle_id"  in frame["value"].keys():
            self.cycle_id = frame["value"]["cycle_id"]

    def get_last_cycle_id(self) -> int:
        return self.cycle_id
    
    def get_all_frames(self) -> list:
        ret = self.frames
        self.frames = []
        return ret
    
    def enable_logging(self, enabled: bool):
        self.protocol_handler.enable_logging(enabled)
        
    def print_devices_list(self) -> None:
        devices_info = "\n\n"
        for d in self.devices.keys():
            devices_info += f"DID: {d}\n"
            for i in self.devices[d]:
                devices_info += f"  - {i}\n"
            devices_info += "\n"

        logger.info(devices_info)
    
    def get_all_devices(self, ble: bool = True, cdc: bool = True) -> dict:
        interfaces_devices = []
        if ble:
            logger.info("Searching for AIRTLS devices on BLE...")
            interfaces_devices += self.ble_interface.find_devices()
        if cdc:
            logger.info("Searching for AIRTLS devices on USB...")
            interfaces_devices += self.serial_interface.find_devices()

        self.devices = {}
        for i in interfaces_devices:
            if i.did not in self.devices.keys():
                self.devices[i.did] = []
            self.devices[i.did].append(i)

        return self.devices

    def get_device(self, did: str, type: str):
        device = None

        try:
            device = next(interface for interface in self.devices[did] if interface.name == type)
        except Exception as e:
            logger.warning(f"No such device as {did} {type}")

        return device