# -----------------------------------------------------------------------------
# File:     abstract_interface.py
# Author:   Daniel Dyringer
# Brief:    Definition of abstract interface and Abstract interface device
# -----------------------------------------------------------------------------

from abc import ABC, abstractmethod
from typing import Callable
import time
import logging
import atexit

logger = logging.getLogger(__name__)

class AbstractInterface(ABC):

    def __init__(
            self, 
            on_data_callback: Callable[[type, int], None] = None
        ) -> None:

        self.callback = on_data_callback
        self.data = bytearray()
        self.last_measurement_ts = time.time()
        self.bytes_in_period = 0

    # Abstract interface
    @abstractmethod
    def find_devices(self) -> list:
        pass

    @abstractmethod
    def _connect(self, device: dict) -> None:
        atexit.register(self.disconnect)
        pass

    @abstractmethod 
    def _disconnect(self) -> None:
        pass

    @abstractmethod 
    def _transmit(self, payload: bytearray) -> None:
        pass
    
    # Common interface
    def connect(self, device: dict) -> None:
        self._connect(device)
        atexit.register(self.disconnect)
        pass

    def disconnect(self) -> None:
        atexit.unregister(self.disconnect)
        self._disconnect()
        self._clean()
        pass

    def transmit(self, payload: bytearray) -> None:
        logger.debug(f"TX: {payload}")
        self._transmit(payload)

    def receive(self, size: int) -> bytearray:
        ret_buffer = bytearray()
        if len(self.data) >= size:
            ret_buffer = self.data[0:size]
            self.data = self.data[size:]
        return ret_buffer
    
    # Private interface
    def _clean(self) -> None:
        self.data = bytearray()
        self.bytes_in_period = 0

    def _on_data(self, payload: bytearray):
        self.bytes_in_period += len(payload)
        now = time.time()
        td = now - self.last_measurement_ts
        if td >= 5.0:
            speed = (self.bytes_in_period) / td
            logger.debug(f"Interface Speed: {speed:.02f} B/s")
            self.bytes_in_period = 0
            self.last_measurement_ts = now

        self.data.extend(payload)
        if self.callback != None:
            self.callback(self, len(self.data))


class InterfaceDevice():

    def __init__(self,
                 did: str,
                 name: str,
                 connection_string: str,
                 interface: AbstractInterface,
                 extra_data=None) -> None:
        
        self.did = did
        self.name = name
        self.connection_string = connection_string
        self.interface: AbstractInterface = interface
        self.extra_data = extra_data

    def __str__(self):
        if self.extra_data != None:
            return f"{self.name}\t{self.connection_string}\t{self.extra_data}"
        else:
            return f"{self.name}\t{self.connection_string}"

    def connect(self):
        self.interface.connect(self.connection_string)

    def disconnect(self):
        logger.warning(f"Disconnecting from {self.name} {self.did}")
        self.interface.disconnect()

    def transmit(self, payload: bytearray) -> None:
        self.interface.transmit(payload)