# -----------------------------------------------------------------------------
# File:     ble.py
# Author:   Daniel Dyringer
# Brief:    Simple implementation of BLE NUS interface
# -----------------------------------------------------------------------------

from bleak import BleakScanner, BleakClient
from threading import Thread
import asyncio
import logging

from .abstract_interface import *

logger = logging.getLogger(__name__)


class InterfaceBLE(AbstractInterface):
    BLE_CHARACTERISTIC_RX = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
    BLE_CHARACTERISTIC_TX = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"
    BLE_MANUFACTURER_ID = 0xFFFF
    BLE_SCAN_TIMEOUT = 2.0

    def __init__(self, on_data_callback: Callable[[bytearray], None] = None) -> None:
        super(InterfaceBLE, self).__init__(on_data_callback)
        self.client = None
        self.loop = asyncio.get_event_loop()
        self.should_disconnect = False
        self.tx_payload = None
        self.rx_thread = None

    def __del__(self):
        self.disconnect()
        self.loop.close()

    def _run_thread(self):
        while True:
            if self.should_disconnect == True:
                self.should_disconnect = False
                self.loop.run_until_complete(self.client.disconnect())
                break
            elif self.tx_payload != None:
                self.loop.run_until_complete(
                    self.client.write_gatt_char(
                        InterfaceBLE.BLE_CHARACTERISTIC_RX, self.tx_payload
                    )
                )
                self.tx_payload = None
            else:
                self.loop.run_until_complete(asyncio.sleep(0.01))

    def find_devices(self) -> list:
        devices = list()

        async def async_find_devices():
            adv_devices = await BleakScanner.discover(
                return_adv=True, cb=dict(), timeout=InterfaceBLE.BLE_SCAN_TIMEOUT
            )
            for d, a in adv_devices.values():
                if (a.local_name is not None and "AIR" in a.local_name) and (
                    a.manufacturer_data is not None
                    and InterfaceBLE.BLE_MANUFACTURER_ID in a.manufacturer_data
                ):
                    address = d.address.split(":")
                    device = InterfaceDevice(
                        did=f"{address[-3]}{address[-2]}{address[-1]}",
                        name=f"{a.local_name}-BLE",
                        connection_string=d.address,
                        interface=self,
                        extra_data=a.manufacturer_data[
                            InterfaceBLE.BLE_MANUFACTURER_ID
                        ],
                    )

                    devices.append(device)

        self.loop.run_until_complete(async_find_devices())

        return devices

    def _connect(self, connection_string: str) -> None:
        async def ble_on_rx(characteristic, data: bytearray):
            self._on_data(data)

        async def async_connect(self: InterfaceBLE, device: dict) -> None:
            self.client = BleakClient(
                connection_string, winrt=dict(use_cached_services=True)
            )
            await self.client.connect()
            logger.info(f"Connected: {self.client.is_connected}")

            # LIST SERVICES/CHARACTERISTICS:
            # for service in self.client.services:
            #     logger.info(Color.BLUE_BOLD_BRIGHT, f"Service: <{service}>")
            #     for char in service.characteristics:
            #         logger.info(Color.CYAN_BOLD_BRIGHT, f"Characteristic: <{char.uuid}>")

            await self.client.start_notify(
                InterfaceBLE.BLE_CHARACTERISTIC_TX, ble_on_rx
            )
            logger.info("Enabled RX notifications")

        self.loop.run_until_complete(async_connect(self, connection_string))
        self.should_disconnect = False
        self.rx_thread = Thread(target=self._run_thread)
        self.rx_thread.start()

    def _disconnect(self) -> None:
        self.should_disconnect = True
        if self.rx_thread != None:
            self.rx_thread.join(timeout=3)

    def _transmit(self, payload: bytearray) -> None:
        while self.tx_payload != None:
            time.sleep(0.1)

        self.tx_payload = payload
