# -----------------------------------------------------------------------------
# File:     cdc.py
# Author:   Daniel Dyringer
# Brief:    Simple implementation of CDC interface
# -----------------------------------------------------------------------------

from threading import Thread
import serial
import time
import logging

from .abstract_interface import *
from third_party.list_ports import list_ports

logger = logging.getLogger(__name__)

class InterfaceSerial(AbstractInterface):

    def __init__(
        self,
        on_data_callback: Callable[[bytearray], None] = None,
    ) -> None:

        super(InterfaceSerial, self).__init__(on_data_callback)
        self.rx_thread = None
        self.serial = None

    def __del__(self):
        self.disconnect()

    def _run_rx_thread(self):
        while self.serial and self.serial.is_open:
            try:
                b = bytearray(self.serial.read(1))
                if len(b) > 0:
                    self._on_data(b)
            except Exception as e:
                pass
            
    def find_devices(self) -> list:
        devices = list()
        ports = list_ports.comports(True)
        for p in ports:
            if p.description == "AIRTLS":

                functions = {0: "PROTOCOL", 2: "SHELL", 4: "TRACE"}
                location = int(p.location.split('.')[-1])
                device = InterfaceDevice(
                    did=p.serial_number[-6:],
                    name=f"{p.description}-USB-{functions[location]}",
                    connection_string=p.device,
                    interface=self,
                )

                devices.append(device)

        return devices

    def _connect(self, connection_string: str) -> None:
        path = connection_string
        self.serial = serial.Serial(baudrate=1000000, port=path, timeout=0.1)
        while self.serial.is_open == False:
            time.sleep(0.1)

        logger.info(f"Connected to {path}")
        self.rx_thread = Thread(target=self._run_rx_thread)
        self.rx_thread.start()

    def _disconnect(self) -> None:
        if self.serial != None:
            self.serial.close()
        if self.rx_thread != None:
            self.rx_thread.join(timeout=3)
        self.serial = None

    def _transmit(self, payload: bytearray) -> None:
        self.serial.write(payload)