# -----------------------------------------------------------------------------
# File:     protocol.py
# Author:   Daniel Dyringer
# Brief:    TLV Protocol parsers
# -----------------------------------------------------------------------------

import rawutil
from abc import ABC, abstractmethod
from .compression import BitStichUnpacker


class AbstractFrame(ABC):

    @staticmethod
    @abstractmethod
    def from_bytes(data: bytearray) -> dict:
        pass

    @staticmethod
    @abstractmethod
    def to_bytes(data: dict) -> bytearray:
        pass


class Header(AbstractFrame):
    structure = "<BH"
    size = rawutil.calcsize(structure)

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        assert len(data) >= Header.size, "Header: Invalid buffer length"

        header = dict()
        header["type"], header["length"] = rawutil.unpack(Header.structure, data)

        return header

    @staticmethod
    def to_bytes(type: int, length: int) -> bytearray:
        return rawutil.pack(Header.structure, type, length)


class TimeOfFlightFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:

        def unpack_rssi(parent: dict, rssi: int):
            parent["rssi_received_by_responder"] = (rssi >> 4) & 0x0F
            parent["rssi_received_by_transmitter"] = (rssi >> 0) & 0x0F

        tmp_rssi = 0x00

        # Unpack constant/format message part
        hdr_format = "<3XH3XB4HBBBB"
        hdr_size = rawutil.calcsize(hdr_format)

        ret = {}
        ret["a2a"] = {}
        ret["a2a"]["a2a_tof"] = [0x0000, 0x0000, 0x0000, 0x0000]
        ret["tag_lists"] = []
        ret["imu"] = {}
        ret["tag_impacts"] = []

        # A2A
        ret["did"],           \
        ret["cycle_id"],      \
        ret["a2a"]["did"],    \
        tmp_rssi,             \
        ret["a2a"]["a2a_tof"][0], \
        ret["a2a"]["a2a_tof"][1], \
        ret["a2a"]["a2a_tof"][2], \
        ret["a2a"]["a2a_tof"][3], \
        ret["imu"]["hit"], \
        ret["imu"]["motion_byte"], \
        ret["imu"]["tag_impact_report_count"], \
        ret["tag_lists_count"] = rawutil.unpack(hdr_format, data[0:hdr_size])

        unpack_rssi(ret["a2a"], tmp_rssi)

        ptr = hdr_size

        # Unpack tag lists
        for i in range(ret["tag_lists_count"]):
            list_format = "<BB"
            list_size = rawutil.calcsize(list_format)

            list = {}
            tmp_list_config = 0x00
            tmp_list_config, list["tag_count"] = rawutil.unpack(list_format, data[ptr:ptr+list_size])
            list["number_of_tofs"] = (tmp_list_config >> 0) & 0x3F
            list["precision"] = (tmp_list_config >> 6) & 0x03
            list["tags"] = []

            ptr += list_size

            # Unpack tags
            for j in range(list["tag_count"]):
                tag_format = "<3XB"
                tag_size = rawutil.calcsize(tag_format)

                tag = {}
                tag["did"], tmp_rssi = rawutil.unpack(tag_format, data[ptr:ptr+tag_size])
                unpack_rssi(tag, tmp_rssi)
                ptr += tag_size

                # Unpack ToFs
                number_of_tofs = list["number_of_tofs"]
                tofs_size = number_of_tofs * 2
                tag["tof"] = rawutil.unpack(f"<{number_of_tofs}H", data[ptr:ptr+tofs_size])
                ptr += tofs_size

                list["tags"].append(tag)

            ret["tag_lists"].append(list)

        impact_report_format = "<3XBB"
        impact_report_format_size = rawutil.calcsize(impact_report_format)
        impact_reports = []

        for _ in range(ret["imu"]["tag_impact_report_count"]):
            did, hit, mb = rawutil.unpack(
                impact_report_format, data[ptr:ptr+impact_report_format_size])
            r = {
                "did": did,
                "hit": hit,
                "motion_byte": mb,
            }
            impact_reports.append(r)

            ptr += impact_report_format_size

        ret["tag_impacts"] = impact_reports

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret

class CompressedDataFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XHB"
        hdr_size = rawutil.calcsize(hdr_format)

        ret = {}

        ret["did"],           \
        ret["cycle_id"],      \
        ret["slot_id"] = rawutil.unpack(hdr_format, data[0:hdr_size])

        ret["serial_data"] = data[hdr_size:].hex()

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret


class SerialDataFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XHB"
        hdr_size = rawutil.calcsize(hdr_format)

        ret = {}

        ret["did"],           \
        ret["cycle_id"],      \
        ret["slot_id"] = rawutil.unpack(hdr_format, data[0:hdr_size])

        ret["serial_data"] = data[hdr_size:].decode("utf-8")

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret


class AccelerometerDataFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XHB"
        hdr_size = rawutil.calcsize(hdr_format)
        ret = {}
        ret["did"],           \
        ret["cycle_id"],      \
        ret["slot_id"] = rawutil.unpack(hdr_format, data[0:hdr_size])
        ret["data"] = []
        ret["data"] = {}
        ret["data"]["x"] = []
        ret["data"]["y"] = []
        ret["data"]["z"] = []

        raw = BitStichUnpacker.unpack_samples(data[hdr_size:], 18)
        for i in range(0, len(raw), 3):
            ret["data"]["x"].append(raw[i + 0])
            ret["data"]["y"].append(raw[i + 1])
            ret["data"]["z"].append(raw[i + 2])

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret

class GyroscopeDataFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XHB"
        hdr_size = rawutil.calcsize(hdr_format)
        ret = {}
        ret["did"],           \
        ret["cycle_id"],      \
        ret["slot_id"] = rawutil.unpack(hdr_format, data[0:hdr_size])
        ret["data"] = {}
        ret["data"]["x"] = []
        ret["data"]["y"] = []
        ret["data"]["z"] = []

        raw = BitStichUnpacker.unpack_samples(data[hdr_size:], 19)
        for i in range(0, len(raw), 3):
            ret["data"]["x"].append(raw[i + 0])
            ret["data"]["y"].append(raw[i + 1])
            ret["data"]["z"].append(raw[i + 2])

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret

class MagnetometerFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XHB"
        hdr_size = rawutil.calcsize(hdr_format)

        ret = {}
        ret["did"],           \
        ret["cycle_id"],      \
        ret["slot_id"] = rawutil.unpack(hdr_format, data[0:hdr_size])

        ret["data"] = {}
        ret["data"]["x"] = []
        ret["data"]["y"] = []
        ret["data"]["z"] = []

        raw = BitStichUnpacker.unpack_samples(data[hdr_size:], 11)
        for i in range(0, len(raw), 3):
            ret["data"]["x"].append(raw[i + 0])
            ret["data"]["y"].append(raw[i + 1])
            ret["data"]["z"].append(raw[i + 2])

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret

class ImpactFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XHB"
        hdr_size = rawutil.calcsize(hdr_format)

        ret = {}
        ret["did"],           \
        ret["cycle_id"],      \
        ret["slot_id"] = rawutil.unpack(hdr_format, data[0:hdr_size])
        ret["hit_type"] = int(data[hdr_size]) & 0x01
        ret["hit_position"] = int(data[hdr_size]) >> 1
        ret["hit_vector"] = int(data[hdr_size + 1])
        if ret["hit_type"] == 1:
            ret["hit_vector"] |= (int(data[hdr_size + 2]) << 8)

        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret


class AdminDistributionInfoFrame(AbstractFrame):

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        # Unpack constant/format message part
        hdr_format = "<3XH3B6X3B4X9s3s2X1B2b1B17B7XB"
        hdr_size = rawutil.calcsize(hdr_format)

        ret = {}
        ret["cycle"] = {}
        ret["bt"] = {}
        ret["uwb"] = {}
        ret["sensor_config"] = {}
        ret["sensor_config"]["magnetoBias"] = {}
        ret["sensor_config"]["magnetoScalar"] = {}
        ret["sensor_config"]["accelerometerBias"] = {}
        ret["sensors_data"] = {}

        ret["did"], \
        ret["cycle_id"], \
        ret["fw_version_major"], \
        ret["fw_version_minor"], \
        ret["hwid"], \
        ret["mac"], \
        ret["cycle"]["device_offset"], \
        ret["cycle"]["frequency"], \
        ret["cycle"]["seat"], \
        ret["serial_number"], \
        ret["name"], \
        ret["group"], \
        ret["license"], \
        ret["bt"]["tx_power_preset"], \
        ret["uwb"]["ch3_antenna_delay"], \
        ret["uwb"]["ch5_antenna_delay"], \
        ret["uwb"]["tx_power_preset"], \
        ret["sensor_config"]["board_config"], \
        ret["sensor_config"]["pressure_calibration"], \
        ret["sensor_config"]["magnetoBias"]["x"], \
        ret["sensor_config"]["magnetoBias"]["y"], \
        ret["sensor_config"]["magnetoBias"]["z"], \
        ret["sensor_config"]["magnetoScalar"]["x"], \
        ret["sensor_config"]["magnetoScalar"]["y"], \
        ret["sensor_config"]["magnetoScalar"]["z"], \
        ret["sensor_config"]["accelerometerBias"]["x"], \
        ret["sensor_config"]["accelerometerBias"]["y"], \
        ret["sensor_config"]["accelerometerBias"]["z"], \
        ret["sensors_data"]["vbat"], \
        ret["sensors_data"]["motion_byte"], \
        ret["sensors_data"]["hit"], \
        ret["sensors_data"]["pressure"], \
        ret["sensors_data"]["heart_rate"], \
        ret["sensors_data"]["mox"], \
        ret["valid"], \
        ret["slow_tags_count"] = rawutil.unpack(hdr_format, data[0:hdr_size])

        ret["name"] = ret["name"].decode("utf-8").strip("\0")
        ret["group"] = ret["group"].decode("utf-8").strip("\0")

        slow_info_format = "<3XBBBB"
        slow_info_size = rawutil.calcsize(slow_info_format)
        slow_info_list = []
        ptr = hdr_size

        for _ in range(ret["slow_tags_count"]):
            slow_info = {}
            slow_info["did"], \
            slow_info["seat"], \
            slow_info["vbat"], \
            slow_info["hwid"], \
            slow_info["motion_byte"] = rawutil.unpack(slow_info_format, data[ptr: ptr + slow_info_size])
        
            ptr += slow_info_size
            slow_info_list.append(slow_info)

        ret["slow_tags"] = slow_info_list
        return ret

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        ret = bytearray()
        # No need to support this
        return ret

class CommandBlinkImmediatelyFrame(AbstractFrame):
    '''
    {
        "anchors": [
            {
                "did": "123456",
                "pattern": 89,
                "brightness" : 1,
            },
            {
                "did": "00000",
                "pattern": 0,
                "brightness" : 0,
            }
            {
                "did": "00000",
                "pattern": 0,
                "brightness" : 0,
            }
        ]
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:

        assert len(data["anchors"]) == 3
        ret = rawutil.pack(
                "<3XB3XB3XB",
                data["anchors"][0]["did"],
                data["anchors"][0]["pattern"] | (data["anchors"][0]["brightness"] << 6),
                data["anchors"][1]["did"],
                data["anchors"][1]["pattern"] | (data["anchors"][1]["brightness"] << 6),
                data["anchors"][2]["did"],
                data["anchors"][2]["pattern"] | (data["anchors"][2]["brightness"] << 6),
            )

        return ret

class CommandBlinkFrame(AbstractFrame):
    '''
    {
        "anchors": [
            {
                "did": "123456",
                "pattern": 89,
                "brightness" : 1,
                "cycle_id": 0,
            },
            {
                "did": "00000",
                "pattern": 0,
                "brightness" : 0,
                "cycle_id": 0,
            }
        ]
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:

        assert len(data["anchors"]) == 2
        ret = rawutil.pack(
                "<3XBH3XBH",
                data["anchors"][0]["did"],
                data["anchors"][0]["pattern"] | (data["anchors"][0]["brightness"] << 6),
                data["anchors"][0]["cycle_id"],
                data["anchors"][1]["did"],
                data["anchors"][1]["pattern"] | (data["anchors"][1]["brightness"] << 6),
                data["anchors"][1]["cycle_id"],
            )

        return ret

class CommandShutdownFrame(AbstractFrame):
    '''
    {
        "anchors": [
            {
                "did": "123456",
            },
            {
                "did": "00000",
            },
            {
                "did": "00000",
            },
            {
                "did": "00000",
            },
        ]
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:

        assert len(data["anchors"]) == 4
        ret = rawutil.pack(
                "<3X3X3X3X",
                data["anchors"][0]["did"],
                data["anchors"][1]["did"],
                data["anchors"][2]["did"],
                data["anchors"][3]["did"],
            )

        return ret

class CommandSetUWBConfigFrame(AbstractFrame):
    '''
    {
        "devices": [
            {
                "did": "123456",
                "ch3": 10,
                "ch5": -1,
                "txPower": 0,
            },
            {
                "did": "000000",
                "ch3": 0,
                "ch5": 0,
                "txPower": 0,
            },
        ]
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        assert len(data["devices"]) == 2
        ret = rawutil.pack(
                "<3XbbB3XbbB",
                data["devices"][0]["did"],
                data["devices"][0]["ch3"],
                data["devices"][0]["ch5"],
                data["devices"][0]["txPower"],
                data["devices"][1]["did"],
                data["devices"][1]["ch3"],
                data["devices"][1]["ch5"],
                data["devices"][1]["txPower"],
            )

        return ret


class CommandSetBoardConfigFrame(AbstractFrame):
    '''
    {
        "devices": [
            {
                "did": "123456",
                "config": 0x01
            },
            {
                "did": "000000",
                "config": 0x00
            },
            {
                "did": "000000",
                "config": 0x00
            },
        ]
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        assert len(data["devices"]) == 3
        ret = rawutil.pack(
                "<3XB3XB3XB",
                data["devices"][0]["did"],
                data["devices"][0]["config"],
                data["devices"][1]["did"],
                data["devices"][1]["config"],
                data["devices"][2]["did"],
                data["devices"][2]["config"],
            )

        return ret


class CommandSetNetworkConfigFrame(AbstractFrame):
    '''
        {
            "devices": [
            {
                "did": "123456",
                "primary_seat": 0x02,
                "frequency": 0x01,
                "allocator_type": 0,
                "device_type": 3,
            },
            {
                "did": "000000",
                "primary_seat": 0x00,
                "frequency": 0x00,
                "allocator_type": 0,
                "device_type": 0,
            },
        }

        primary_seat - absolute seat number
        frequency:
            0 - Unknown      (should not be used),
            1 - Random Tag   (not implemented),
            2 - VerySlow Tag (not implemented),
            3 - Slow Tag     (not implemented),

            4 - 25Hz,
            5 - 50Hz,
            6 - 75Hz,   (not implemented)
            7 - 100Hz,  (not implemented)
            8 - 125Hz,  (not implemented)
            9 - 150Hz,

        allocator_type:
            0 - static
            1 - dynamic strict
            2 - dynamic relaxed
        
        device_type:
            0 - None    (should not be used),
            1 - Dummy   (should not be used),
            2 - Anchor,
            3 - Tag,
            4 - Sniffer,
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        assert len(data["devices"]) == 2

        def validate(index: int):
            if data["devices"][index]["did"] != "000000":
                assert int(data["devices"][index]["device_type"]) in [2, 3, 4]
                assert int(data["devices"][index]["allocator_type"]) in [0, 1, 2]
                assert int(data["devices"][index]["frequency"]) in [1, 2, 3, 4, 5, 9]
        
        validate(0)
        validate(1)

        def createConfigByte(index: int) -> int:
            config = int(data["devices"][index]["allocator_type"] & 0x03)
            config |= (int(data["devices"][index]["device_type"]) & 0x07) << 2
            return config

        ret = rawutil.pack(
                "<3XBBB3XBBB",
                data["devices"][0]["did"],
                data["devices"][0]["primary_seat"],
                data["devices"][0]["frequency"],
                createConfigByte(0),
                data["devices"][1]["did"],
                data["devices"][1]["primary_seat"],
                data["devices"][1]["frequency"],
                createConfigByte(1)
            )

        assert(len(ret) == 12)

        return ret


class CommandSetRandomPeriod(AbstractFrame):
    '''
    {
        "devices": [
            {
                "did": "123456",
                "period": 0x0101
            },
            {
                "did": "000000",
                "period": 0x0000
            },
        ]

        period - random frequency tag active period in seconds
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        assert len(data["devices"]) == 2
        
        def validate(index: int):
            if data["devices"][index]["did"] != "000000":
                assert int(data["devices"][index]["period"]) >= 30
        
        validate(0)
        validate(1)

        ret = rawutil.pack(
                "<3XH3XHBB",
                data["devices"][0]["did"],
                data["devices"][0]["period"],
                data["devices"][1]["did"],
                data["devices"][1]["period"],
                0,
                0
            )
        
        assert(len(ret) == 12)

        return ret

class CommandSetNetworkStartDelay(AbstractFrame):
    '''
    {
        "devices": [
            {
                "did": "123456",
                "delay": 0
            },
            {
                "did": "000000",
                "delay": 255,
            },
            {
                "did": "000000",
                "delay": 255,
            },
        ]

        delay - network start delay for an anchor in seconds
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:
        assert len(data["devices"]) == 3
        
        def validate(index: int):
            if data["devices"][index]["did"] != "000000":
                assert int(data["devices"][index]["delay"]) <= 255
        
        for i in range(3):
            validate(i)

        ret = rawutil.pack(
                "<3XB3XB3XB",
                data["devices"][0]["did"],
                data["devices"][0]["delay"],
                data["devices"][1]["did"],
                data["devices"][1]["delay"],
                data["devices"][2]["did"],
                data["devices"][2]["delay"],
            )
        
        assert(len(ret) == 12)

        return ret


class CommandFactoryResetFrame(AbstractFrame):
    '''
    {
        "anchors": [
            {
                "did": "123456",
            },
            {
                "did": "00000",
            },
            {
                "did": "00000",
            },
            {
                "did": "00000",
            },
        ]
    }
    '''

    @staticmethod
    def from_bytes(data: bytearray) -> dict:
        return {}

    @staticmethod
    def to_bytes(data: dict) -> bytearray:

        assert len(data["anchors"]) == 4
        ret = rawutil.pack(
                "<3X3X3X3X",
                data["anchors"][0]["did"],
                data["anchors"][1]["did"],
                data["anchors"][2]["did"],
                data["anchors"][3]["did"],
            )

        return ret


class FrameType():

    frame_classes = {
            1 :  TimeOfFlightFrame,
            2 :  SerialDataFrame,
            3 :  AdminDistributionInfoFrame,
            4 :  CompressedDataFrame,
            5 :  AccelerometerDataFrame,
            6 :  GyroscopeDataFrame,
            7 :  MagnetometerFrame,
            8 :  ImpactFrame,

            139: CommandSetUWBConfigFrame,
            140: CommandShutdownFrame,
            141: CommandSetBoardConfigFrame,
            142: CommandSetNetworkConfigFrame,
            143: CommandSetRandomPeriod,
            144: CommandSetNetworkStartDelay,
            198: CommandFactoryResetFrame,
            200: CommandBlinkImmediatelyFrame,
            201: CommandBlinkFrame
    }

    @staticmethod
    def get_class(frame_type: int) -> type:
        return FrameType.frame_classes[frame_type]

    def get_type(frame_class: type) -> int:
        ret = {i for i in FrameType.frame_classes if FrameType.frame_classes[i]==frame_class}
        return  list(ret)[0]


def protocol_helper_createBinaryFrame(frame_class: type | int, data: dict)-> bytearray:
    message_type = 0
    message_class = None

    if type(frame_class) ==  int:
        message_type = frame_class
        message_class = FrameType.get_class(frame_class)
    else:
        message_type = FrameType.get_type(frame_class)
        message_class = frame_class

    value = message_class.to_bytes(data)
    hdr = Header.to_bytes(message_type, len(value))
    return hdr + value
