# -----------------------------------------------------------------------------
# File:     protocol_builder.py
# Author:   Daniel Dyringer
# Brief:    AIRTLS Protocol builder - creates TLV protocol frames from byte stream
# -----------------------------------------------------------------------------

from typing import Callable
import logging

from .interface.abstract_interface import *
from .protocol  import *

logger = logging.getLogger(__name__)

class ProtocolBuilder():
    
    def __init__(self, on_frame_callback: Callable[[dict], None]) -> None:
        self.frame_data = bytearray()
        self.header = None
        self.frame = None
        self.logging = False
        self.callback = on_frame_callback
        
    def _clean(self):
        self.frame_data = bytearray()
        self.header = None
        self.frame = None
    
    def _log_frame(self):
        logger.info({"header": self.header, "value": self.frame})

    def enable_logging(self, enabled: bool):
        self.logging = enabled

    def on_data(self, interface: AbstractInterface, size: int):
        HDR_LEN = 3
        reaming_size = size
        while True:
            
            if (len(self.frame_data) == 0):
                if (reaming_size >= HDR_LEN):
                    self.frame_data.extend(interface.receive(HDR_LEN))
                    try:
                        self.header = Header.from_bytes(self.frame_data[0:HDR_LEN])
                        reaming_size = reaming_size - HDR_LEN
                    except Exception as e:
                        logger.error(f"Header error: {e}")
                        interface.receive(size)
                        self._clean()
                else:
                    break

            if (len(self.frame_data) == HDR_LEN) and (self.header != None):
                
                if (self.header["type"] not in FrameType.frame_classes.keys()) or (self.header["length"] > 1000):
                    interface.receive(size)
                    self._clean()

                if (reaming_size >= self.header["length"]):
                    self.frame_data.extend(interface.receive(self.header["length"]))
                    reaming_size -= self.header["length"]
                    assert len(self.frame_data) == (HDR_LEN + self.header["length"])

                    try:
                        frame_class: AbstractFrame = FrameType.get_class(self.header["type"])
                        self.frame = frame_class.from_bytes(self.frame_data[HDR_LEN:])
                        if self.logging:
                            self._log_frame()
                        if self.callback != None:
                            self.callback({"header": self.header, "value": self.frame})

                    except Exception as e:
                        logger.error(f"Payload error: {frame_class} {e}")
                        interface.receive(size)
                    finally:
                        self._clean()
                else:
                    break